/*
 * Copyright (C) 2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ActivityState.h"

#include <wtf/text/TextStream.h>

namespace WebCore {

TextStream& operator<<(TextStream& ts, OptionSet<ActivityState> flags)
{
    bool didAppend = false;

    auto appendIf = [&](auto flag, auto message) {
        if (!flags.contains(flag))
            return;
        if (didAppend)
            ts << ", "_s;
        ts << message;
        didAppend = true;
    };
    
    appendIf(ActivityState::WindowIsActive, "active window");
    appendIf(ActivityState::IsFocused, "focused");
    appendIf(ActivityState::IsVisible, "visible");
    appendIf(ActivityState::IsVisibleOrOccluded, "visible or occluded");
    appendIf(ActivityState::IsInWindow, "in-window");
    appendIf(ActivityState::IsVisuallyIdle, "visually idle");
    appendIf(ActivityState::IsAudible, "audible");
    appendIf(ActivityState::IsLoading, "loading");
    appendIf(ActivityState::IsCapturingMedia, "capturing media");
    appendIf(ActivityState::IsConnectedToHardwareConsole, "attached to hardware console");

    return ts;
}

} // namespace WebCore
