/*
 * Copyright 2023 The Chromium Authors. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(WEBGL)

#include "WebGLObject.h"

namespace WTF {
class AbstractLocker;
}

namespace WebCore {

class WebGLTimerQueryEXT final : public WebGLObject {
public:
    static RefPtr<WebGLTimerQueryEXT> create(WebGLRenderingContextBase&);
    virtual ~WebGLTimerQueryEXT();

    bool isResultAvailable() const { return m_isResultAvailable; }
    void makeResultAvailable() { m_isResultAvailable = true; }

    void setTarget(GCGLenum target) { m_target = target; }
    GCGLenum target() const { return m_target; }
    void didBind(GCGLenum target) { setTarget(target); }
    bool isUsable() const { return object() && !isDeleted(); }
    bool isInitialized() const { return true; }

private:
    WebGLTimerQueryEXT(WebGLRenderingContextBase&, PlatformGLObject);
    void deleteObjectImpl(const AbstractLocker&, GraphicsContextGL*, PlatformGLObject) override;

    bool m_isResultAvailable { false };
    GCGLenum m_target { 0 };
};

} // namespace WebCore

#endif
