/*
 * Copyright (C) 2006, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Samuel Weinig <sam.weinig@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

[
    Exposed=Window
] interface HTMLButtonElement : HTMLElement {
    [CEReactions=NotNeeded, Reflect] attribute boolean disabled;
    [ImplementedAs=formForBindings] readonly attribute HTMLFormElement? form;
    [CEReactions=NotNeeded, ReflectSetter] attribute [AtomString] USVString formAction;

    [CEReactions=NotNeeded, ReflectSetter] attribute [AtomString] DOMString formEnctype;
    [CEReactions=NotNeeded, ReflectSetter] attribute [AtomString] DOMString formMethod;
    [CEReactions=NotNeeded, ReflectSetter] attribute [AtomString] DOMString type;

    [CEReactions=NotNeeded, Reflect] attribute boolean formNoValidate;
    [CEReactions=NotNeeded, Reflect] attribute DOMString formTarget;
    [CEReactions=NotNeeded, Reflect] attribute DOMString name;
    [CEReactions=NotNeeded, Reflect] attribute DOMString value;

    // Command Invokers
    [CEReactions=NotNeeded, Reflect="commandfor", EnabledBySetting=CommandAttributesEnabled] attribute Element? commandForElement;
    [CEReactions=NotNeeded, EnabledBySetting=CommandAttributesEnabled, ReflectSetter] attribute [AtomString] DOMString command;

    readonly attribute boolean willValidate;
    readonly attribute ValidityState validity;
    readonly attribute DOMString validationMessage;
    boolean checkValidity();
    [EnabledBySetting=InteractiveFormValidationEnabled] boolean reportValidity();
    undefined setCustomValidity(DOMString error);

    // FIXME: `labels` should not be nullable.
    readonly attribute NodeList? labels;
};

HTMLButtonElement includes PopoverInvokerElement;
