/*
 * Copyright (C) 2022, 2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "CSSStyleSheet.h"
#include "JSObservableArray.h"

namespace WebCore {

class ContainerNode;
template<typename> class ExceptionOr;

class CSSStyleSheetObservableArray : public JSC::ObservableArray {
public:
    static Ref<CSSStyleSheetObservableArray> create(ContainerNode& treeScope);

    ExceptionOr<void> setSheets(Vector<Ref<CSSStyleSheet>>&&);
    const Vector<Ref<CSSStyleSheet>>& sheets() const { return m_sheets; }

private:
    explicit CSSStyleSheetObservableArray(ContainerNode& treeScope);

    std::optional<Exception> shouldThrowWhenAddingSheet(const CSSStyleSheet&) const;

    // JSC::ObservableArray
    bool setValueAt(JSC::JSGlobalObject*, unsigned index, JSC::JSValue) final;
    void removeLast() final;
    JSC::JSValue valueAt(JSC::JSGlobalObject*, unsigned index) const final;
    unsigned length() const final { return m_sheets.size(); }
    void shrinkTo(unsigned) final;

    TreeScope* treeScope() const;

    void didAddSheet(CSSStyleSheet&);
    void willRemoveSheet(CSSStyleSheet&);

    WeakPtr<ContainerNode, WeakPtrImplWithEventTargetData> m_treeScope;
    Vector<Ref<CSSStyleSheet>> m_sheets;
};

} // namespace WebCore
