/*
 * Copyright (C) 2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(MEDIA_STREAM)

#include <optional>

namespace WebCore {

enum class FillLightMode : uint8_t;

struct PhotoSettings {
    PhotoSettings isolatedCopy() const { return *this; }

    std::optional<FillLightMode> fillLightMode;
    std::optional<double> imageHeight;
    std::optional<double> imageWidth;
    std::optional<bool> redEyeReduction;
};

inline bool operator==(const PhotoSettings& a, const PhotoSettings& b)
{
    if (!!a.fillLightMode != !!b.fillLightMode)
        return false;
    if (a.fillLightMode && *a.fillLightMode != *b.fillLightMode)
        return false;

    if (!!a.imageHeight != !!b.imageHeight)
        return false;
    if (a.imageHeight && *a.imageHeight != *b.imageHeight)
        return false;

    if (!!a.imageWidth != !!b.imageWidth)
        return false;
    if (a.imageWidth && *a.imageWidth != *b.imageWidth)
        return false;

    if (!!a.redEyeReduction != !!b.redEyeReduction)
        return false;
    if (a.redEyeReduction && *a.redEyeReduction != *b.redEyeReduction)
        return false;

    return true;
}

}

#endif
