/*
 * Copyright (C) 2014, 2015 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include <JavaScriptCore/CallVariant.h>

namespace JSC {

class CallEdge {
public:
    CallEdge();
    CallEdge(CallVariant, uint32_t);
    
    bool operator!() const { return !m_callee; }
    
    CallVariant callee() const { return m_callee; }
    uint32_t count() const { return m_count; }
    
    CallEdge despecifiedClosure() const
    {
        return CallEdge(m_callee.despecifiedClosure(), m_count);
    }
    
    void dump(PrintStream&) const;
    
private:
    CallVariant m_callee;
    uint32_t m_count;
};

inline CallEdge::CallEdge(CallVariant callee, uint32_t count)
    : m_callee(callee)
    , m_count(count)
{
}

inline CallEdge::CallEdge()
    : CallEdge(CallVariant(), 0)
{
}

typedef Vector<CallEdge, 1> CallEdgeList;

} // namespace JSC
