/*
 * Copyright (C) 2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <JavaScriptCore/JSExportMacros.h>
#include <utility>
#include <wtf/Forward.h>
#include <wtf/TriState.h>

namespace JSC {

class CallFrame;
class VM;

enum class SourceTaintedOrigin : uint8_t {
    Untainted = 0,
    IndirectlyTaintedByHistory, // Means the VM saw some tainted code this event loop turn but no such code was on the stack when this source was created.
    IndirectlyTainted,
    KnownTainted,
};

inline TriState taintednessToTriState(SourceTaintedOrigin origin)
{
    if (origin == SourceTaintedOrigin::Untainted)
        return TriState::False;
    if (origin == SourceTaintedOrigin::KnownTainted)
        return TriState::True;
    return TriState::Indeterminate;
}


JS_EXPORT_PRIVATE std::pair<SourceTaintedOrigin, URL> sourceTaintedOriginFromStack(VM&, CallFrame*);
JS_EXPORT_PRIVATE SourceTaintedOrigin computeNewSourceTaintedOriginFromStack(VM&, CallFrame*);

JS_EXPORT_PRIVATE String sourceTaintedOriginToString(SourceTaintedOrigin taintedness);

}

