/*
 * Copyright (C) 2025 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <WebCore/HTMLNames.h>
#include <WebCore/Node.h>
#include <WebCore/RenderObject.h>

namespace WebCore {

inline Node& RenderObject::nodeForNonAnonymous() const
{
    ASSERT(!isAnonymous());
    return m_node.get();
}

inline Node* RenderObject::nonPseudoNode() const
{
    if (isPseudoElement())
        return nullptr;
    return node();
}

inline Node* RenderObject::node() const
{
    if (isAnonymous())
        return nullptr;
    return m_node.ptr();
}

inline RefPtr<Node> RenderObject::protectedNode() const
{
    return node();
}

inline bool RenderObject::isBody() const { return node() && node()->hasTagName(HTMLNames::bodyTag); }
inline bool RenderObject::isHR() const { return node() && node()->hasTagName(HTMLNames::hrTag); }
inline bool RenderObject::isPseudoElement() const { return node() && node()->isPseudoElement(); }

} // namespace WebCore
