/*
 * Copyright (C) 2019-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "MIMETypeCache.h"

#include "ContentType.h"
#include <wtf/TZoneMallocInlines.h>

namespace WebCore {

WTF_MAKE_TZONE_ALLOCATED_IMPL(MIMETypeCache);

void MIMETypeCache::ensureSupportedTypes()
{
    Locker locker { m_supportedTypesLock };
    if (!m_supportedTypes) {
        m_supportedTypes = HashSet<String> { };
        initializeCache(*m_supportedTypes);
    }
}

HashSet<String> MIMETypeCache::supportedTypes()
{
    ensureSupportedTypes();

    Locker locker { m_supportedTypesLock };
    return *m_supportedTypes;
}

bool MIMETypeCache::supportsContainerType(const String& containerType)
{
    if (!isAvailable() || containerType.isEmpty())
        return false;

    if (isUnsupportedContainerType(containerType))
        return false;

    if (isStaticContainerType(containerType))
        return true;

    ensureSupportedTypes();

    Locker locker { m_supportedTypesLock };
    return m_supportedTypes->contains(containerType);
}

MediaPlayerEnums::SupportsType MIMETypeCache::canDecodeType(const String& mimeType)
{
    if (mimeType.isEmpty())
        return MediaPlayerEnums::SupportsType::IsNotSupported;

    if (auto supports = getCachedResult(mimeType))
        return *supports;

    auto result = MediaPlayerEnums::SupportsType::IsNotSupported;
    do {
        if (!isAvailable() || mimeType.isEmpty())
            break;

        auto contentType = ContentType { mimeType };
        auto containerType = contentType.containerType();
        if (!supportsContainerType(containerType))
            break;

        if (contentType.codecs().isEmpty()) {
            result = MediaPlayerEnums::SupportsType::MayBeSupported;
            break;
        }

        if (shouldOverrideExtendedType(contentType)) {
            result = MediaPlayerEnums::SupportsType::IsSupported;
            break;
        }

        if (canDecodeExtendedType(contentType))
            result = MediaPlayerEnums::SupportsType::IsSupported;

    } while (0);

    addCachedResult(mimeType, result);

    return result;
}

void MIMETypeCache::addSupportedTypes(const Vector<String>& newTypes)
{
    ensureSupportedTypes();

    Locker locker { m_supportedTypesLock };
    for (auto& type : newTypes)
        m_supportedTypes->add(type);
}

std::optional<MediaPlayerEnums::SupportsType> MIMETypeCache::getCachedResult(const String& mimeType) const
{
    Locker locker { m_cachedResultsLock };
    if (m_cachedResults)
        return m_cachedResults->getOptional(mimeType);
    return std::nullopt;
}

void MIMETypeCache::addCachedResult(const String& mimeType, MediaPlayerEnums::SupportsType result)
{
    Locker locker { m_cachedResultsLock };
    if (!m_cachedResults)
        m_cachedResults = HashMap<String, MediaPlayerEnums::SupportsType>();
    m_cachedResults->add(mimeType, result);
}

bool MIMETypeCache::isStaticContainerType(StringView)
{
    return false;
}

bool MIMETypeCache::isUnsupportedContainerType(const String&)
{
    return false;
}

bool MIMETypeCache::isAvailable() const
{
    return true;
}

bool MIMETypeCache::isEmpty() const
{
    Locker locker { m_supportedTypesLock };
    return m_supportedTypes && m_supportedTypes->isEmpty();
}

void MIMETypeCache::initializeCache(HashSet<String>&)
{
}

bool MIMETypeCache::canDecodeExtendedType(const ContentType&)
{
    return false;
}

bool MIMETypeCache::shouldOverrideExtendedType(const ContentType& type)
{
    ASSERT(canDecodeType(type.containerType()) != MediaPlayerEnums::SupportsType::IsNotSupported);

    // Some sites (e.g. Modernizr) use 'audio/mpeg; codecs="mp3"' even though
    // it is not RFC 3003 compliant.
    if (equalLettersIgnoringASCIICase(type.containerType(), "audio/mpeg"_s)) {
        auto codecs = type.codecs();
        return codecs.size() == 1 && codecs[0] == "mp3"_s;
    }

    return false;
}

}
