/*
 * Copyright (C) 2025 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

.time-control {
    display: flex;
    column-gap: 5px;
    container: time-control-container / size;
    justify-content: space-between;
}

.time-control .scrubber {
    flex-grow: 1;
}

/* Support displaying the time labels above or below the scrubber, rather than side-by-side: */
.time-control.above,
.time-control.below {

    flex-wrap: wrap;
    .scrubber { width: 100%; }
}
.time-control.above {
    .scrubber { order: 2; }
    .time-label { order: 1; }
}
.time-control.below {
    .scrubber { order: 1; }
    .time-label { order: 2; }
}

/* Support toggling the duration and remaining time labels: */
.time-control.duration > #time-label-remaining { display: none; }
.time-control.remaining > #time-label-duration { display: none; }

@container time-control-container (width < 160px) {
    .time-control > #time-label-duration,
    .time-control > #time-label-elapsed {
        display: none;
    }
    /* Override the duration/remaining toggle when in small control mode: */
    .time-control.duration > #time-label-remaining { display: revert; }
}
