/*
 * Copyright (C) 2022 Apple Inc. All rights reserved.
 * Copyright (C) 2023 Igalia S.L
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */


// FIXME: Support Serializable and Transferable.
[
    Conditional=WEB_CODECS,
    EnabledBySetting=WebCodecsAudioEnabled,
    Exposed=(Window,DedicatedWorker),
    InterfaceName=AudioData,
    ReportExtraMemoryCost,
] interface WebCodecsAudioData {
    [CallWith=CurrentScriptExecutionContext] constructor(AudioDataInit init);

    readonly attribute AudioSampleFormat? format;
    readonly attribute float sampleRate;
    readonly attribute unsigned long numberOfFrames;
    readonly attribute unsigned long numberOfChannels;
    readonly attribute unsigned long long? duration;
    readonly attribute long long timestamp;

    unsigned long allocationSize(optional AudioDataCopyToOptions options = {});
    undefined copyTo([AllowShared] (ArrayBufferView or ArrayBuffer) destination, AudioDataCopyToOptions options);
    [CallWith=CurrentScriptExecutionContext] WebCodecsAudioData clone();
    undefined close();
};

[
    Conditional=WEB_CODECS,
    LegacyNativeDictionaryRequiredInterfaceNullability,
] dictionary AudioDataInit {
    required AudioSampleFormat format;
    required float sampleRate;
    required [EnforceRange] long long timestamp;
    required (ArrayBufferView or ArrayBuffer) data;
    required [EnforceRange] unsigned long numberOfFrames;
    required [EnforceRange] unsigned long numberOfChannels;
};

[
    Conditional=WEB_CODECS,
    LegacyNativeDictionaryRequiredInterfaceNullability,
] dictionary AudioDataCopyToOptions {
    [EnforceRange] unsigned long planeIndex;
    [EnforceRange] unsigned long frameOffset;
    [EnforceRange] unsigned long frameCount;
    AudioSampleFormat format;
};
