/*
 *  Copyright (C) 2025 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <wtf/RefCountedWithInlineWeakPtr.h>
#include <wtf/UniquelyOwnedPtr.h>

namespace WTF {

// Use CheckedPtr for strong references and InlineWeakPtr for weak references.
// When a CheckedPtr outlives a UniquelyOwnedPtr:
//     * In debug builds, the CheckedPtr ASSERTs
//     * In release builds, the CheckedPtr keeps the pointed-to object alive.
template<typename T>
class UniquelyOwned : private RefCountedWithInlineWeakPtr<T> {
    friend class RefCountedWithInlineWeakPtr<T>;
    template<typename U> friend UniquelyOwnedPtr<U> adoptUniquelyOwned(U*);
public:
    // UniquelyOwned interface
    void disown() const
    {
        this->refCountDebugger().willDelete();
        this->deref();
    }

    // CheckedPtr interface
    void incrementCheckedPtrCount() const
    {
        this->ref();
    }

    void decrementCheckedPtrCount() const
    {
        // If this ASSERT fires, it means that a CheckedPtr outlived a UniquelyOwnedPtr.
        // Note: There's no security risk here -- CheckedPtr is a strong pointer in release builds.
        ASSERT_WITH_SECURITY_IMPLICATION(!this->refCountDebugger().deletionHasBegun());
        this->deref();
    }

    uint32_t checkedPtrCount() const
    {
        // If this ASSERT fires, it means that a CheckedPtr outlived a UniquelyOwnedPtr.
        // Note: There's no security risk here -- CheckedPtr is a strong pointer in release builds.
        ASSERT_WITH_SECURITY_IMPLICATION(!this->refCountDebugger().deletionHasBegun());
        return this->refCount() - 1; // One refcount belongs to our owner
    }

    uint32_t checkedPtrCountWithoutThreadCheck() const
    {
        // If this ASSERT fires, it means that a CheckedPtr outlived a UniquelyOwnedPtr.
        // Note: There's no security risk here -- CheckedPtr is a strong pointer in release builds.
        ASSERT_WITH_SECURITY_IMPLICATION(!this->refCountDebugger().deletionHasBegun());
        return this->refCount() - 1; // One refcount belongs to our owner
    }

    // InlineWeakPtr interface
    using RefCountedWithInlineWeakPtr<T>::weakRef;
    using RefCountedWithInlineWeakPtr<T>::weakDeref;
    using RefCountedWithInlineWeakPtr<T>::refCount;
};

} // namespace WTF

using WTF::UniquelyOwned;
