/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 * Copyright (C) 2013-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

enum MediaStreamTrackState { "live", "ended" };

// https://w3c.github.io/mediacapture-main/#dom-mediastreamtrack
[
    ActiveDOMObject,
    Conditional=MEDIA_STREAM,
    ExportToWrappedFunction,
    PrivateIdentifier,
    PublicIdentifier,
    Exposed=(Window,DedicatedWorker)
] interface MediaStreamTrack : EventTarget {
    readonly attribute DOMString kind;
    readonly attribute DOMString id;
    readonly attribute DOMString label;
    attribute boolean enabled;
    attribute DOMString contentHint;
    [ImplementedAs=mutedForBindings] readonly attribute boolean muted;
    attribute EventHandler onmute;
    attribute EventHandler onunmute;
    attribute EventHandler onconfigurationchange;
    readonly attribute MediaStreamTrackState readyState;
    attribute EventHandler onended;

    // FIXME: `clone` should not return a nullable type.
    MediaStreamTrack? clone();
    [ImplementedAs=stopTrack] undefined stop();

    MediaTrackCapabilities getCapabilities();
    MediaTrackConstraints getConstraints();
    MediaTrackSettings getSettings();
    Promise<undefined> applyConstraints(optional MediaTrackConstraints constraints);
};

[
    Conditional=MEDIA_STREAM,
    JSGenerateToJSObject,
    LegacyNativeDictionaryRequiredInterfaceNullability,
] dictionary MediaTrackSettings {
    long width;
    long height;
    double aspectRatio;
    double frameRate;
    DOMString facingMode;
    double volume;
    long sampleRate;
    long sampleSize;
    boolean echoCancellation;
    // FIXME 169871: add latency
    // FIXME 169871: add channelCount
    DOMString deviceId;
    DOMString groupId;
    DOMString displaySurface;

    DOMString whiteBalanceMode;
    // FIXME: add exposureMode;
    // FIXME: add focusMode;
    // FIXME: add pointsOfInterest;

    // FIXME: add exposureCompensation;
    // FIXME: add exposureTime;
    // FIXME: add colorTemperature;
    // FIXME: add iso;

    // FIXME: add brightness;
    // FIXME: add contrast;
    // FIXME: add saturation;
    // FIXME: add sharpness;

    // FIXME: add focusDistance;
    double zoom;

    boolean torch;

    boolean backgroundBlur;
    boolean powerEfficient;
};
