// This file is automatically generated from CSSProperties.json by the process-css-properties.py script. Do not edit it.

#pragma once

#ifndef RENDER_STYLE_PROPERTIES_GETTERS_INLINES_INCLUDE_TRAP
#error "Please do not include this file anywhere except from RenderStyle+GettersInlines.h."
#endif

#include <WebCore/RenderStyleBase+GettersInlines.h>
#include <WebCore/StyleComputedStyle+GettersInlines.h>

namespace WebCore {

inline const Style::Color& RenderStyleProperties::testColor() const
{
    return m_computedStyle.testColor();
}

inline decltype(auto) RenderStyleProperties::testColorResolver() const
{
    return m_computedStyle.testColorResolver();
}

inline WebCore::Color RenderStyleProperties::testColorResolvingCurrentColor() const
{
    return m_computedStyle.testColorResolvingCurrentColor();
}

inline WebCore::Color RenderStyleProperties::testColorResolvingCurrentColorApplyingColorFilter() const
{
    return m_computedStyle.testColorResolvingCurrentColorApplyingColorFilter();
}

inline WebCore::Color RenderStyleProperties::visitedDependentTestColor(OptionSet<PaintBehavior> paintBehavior) const
{
    return m_computedStyle.visitedDependentTestColor(paintBehavior);
}

inline WebCore::Color RenderStyleProperties::visitedDependentTestColorApplyingColorFilter(OptionSet<PaintBehavior> paintBehavior) const
{
    return m_computedStyle.visitedDependentTestColorApplyingColorFilter(paintBehavior);
}

inline const Style::Color& RenderStyleProperties::testColorAllowsTypesAbsolute() const
{
    return m_computedStyle.testColorAllowsTypesAbsolute();
}

inline decltype(auto) RenderStyleProperties::testColorAllowsTypesAbsoluteResolver() const
{
    return m_computedStyle.testColorAllowsTypesAbsoluteResolver();
}

inline WebCore::Color RenderStyleProperties::testColorAllowsTypesAbsoluteResolvingCurrentColor() const
{
    return m_computedStyle.testColorAllowsTypesAbsoluteResolvingCurrentColor();
}

inline WebCore::Color RenderStyleProperties::testColorAllowsTypesAbsoluteResolvingCurrentColorApplyingColorFilter() const
{
    return m_computedStyle.testColorAllowsTypesAbsoluteResolvingCurrentColorApplyingColorFilter();
}

inline WebCore::Color RenderStyleProperties::visitedDependentTestColorAllowsTypesAbsolute(OptionSet<PaintBehavior> paintBehavior) const
{
    return m_computedStyle.visitedDependentTestColorAllowsTypesAbsolute(paintBehavior);
}

inline WebCore::Color RenderStyleProperties::visitedDependentTestColorAllowsTypesAbsoluteApplyingColorFilter(OptionSet<PaintBehavior> paintBehavior) const
{
    return m_computedStyle.visitedDependentTestColorAllowsTypesAbsoluteApplyingColorFilter(paintBehavior);
}

inline const Style::Color& RenderStyleProperties::testColorPropertyWithVisitedLinkSupport() const
{
    return m_computedStyle.testColorPropertyWithVisitedLinkSupport();
}

inline const Style::Color& RenderStyleProperties::visitedLinkTestColorPropertyWithVisitedLinkSupport() const
{
    return m_computedStyle.visitedLinkTestColorPropertyWithVisitedLinkSupport();
}

inline decltype(auto) RenderStyleProperties::testColorPropertyWithVisitedLinkSupportResolver() const
{
    return m_computedStyle.testColorPropertyWithVisitedLinkSupportResolver();
}

inline WebCore::Color RenderStyleProperties::testColorPropertyWithVisitedLinkSupportResolvingCurrentColor() const
{
    return m_computedStyle.testColorPropertyWithVisitedLinkSupportResolvingCurrentColor();
}

inline WebCore::Color RenderStyleProperties::testColorPropertyWithVisitedLinkSupportResolvingCurrentColorApplyingColorFilter() const
{
    return m_computedStyle.testColorPropertyWithVisitedLinkSupportResolvingCurrentColorApplyingColorFilter();
}

inline WebCore::Color RenderStyleProperties::visitedLinkTestColorPropertyWithVisitedLinkSupportResolvingCurrentColor() const
{
    return m_computedStyle.visitedLinkTestColorPropertyWithVisitedLinkSupportResolvingCurrentColor();
}

inline WebCore::Color RenderStyleProperties::visitedLinkTestColorPropertyWithVisitedLinkSupportResolvingCurrentColorApplyingColorFilter() const
{
    return m_computedStyle.visitedLinkTestColorPropertyWithVisitedLinkSupportResolvingCurrentColorApplyingColorFilter();
}

inline WebCore::Color RenderStyleProperties::visitedDependentTestColorPropertyWithVisitedLinkSupport(OptionSet<PaintBehavior> paintBehavior) const
{
    return m_computedStyle.visitedDependentTestColorPropertyWithVisitedLinkSupport(paintBehavior);
}

inline WebCore::Color RenderStyleProperties::visitedDependentTestColorPropertyWithVisitedLinkSupportApplyingColorFilter(OptionSet<PaintBehavior> paintBehavior) const
{
    return m_computedStyle.visitedDependentTestColorPropertyWithVisitedLinkSupportApplyingColorFilter(paintBehavior);
}

inline Style::Number<> RenderStyleProperties::testRenderStyleHasExplicitlySetPolicyAllAuthorOrigin() const
{
    return m_computedStyle.testRenderStyleHasExplicitlySetPolicyAllAuthorOrigin();
}

inline bool RenderStyleProperties::hasExplicitlySetTestRenderStyleHasExplicitlySetPolicyAllAuthorOrigin() const
{
    return m_computedStyle.hasExplicitlySetTestRenderStyleHasExplicitlySetPolicyAllAuthorOrigin();
}

inline Style::Number<> RenderStyleProperties::testRenderStyleHasExplicitlySetPolicyAllBorderRadius() const
{
    return m_computedStyle.testRenderStyleHasExplicitlySetPolicyAllBorderRadius();
}

inline bool RenderStyleProperties::hasExplicitlySetTestRenderStyleHasExplicitlySetPolicyAllBorderRadius() const
{
    return m_computedStyle.hasExplicitlySetTestRenderStyleHasExplicitlySetPolicyAllBorderRadius();
}

inline Style::Number<> RenderStyleProperties::testRenderStyleHasExplicitlySetPolicyValueOnly() const
{
    return m_computedStyle.testRenderStyleHasExplicitlySetPolicyValueOnly();
}

inline bool RenderStyleProperties::hasExplicitlySetTestRenderStyleHasExplicitlySetPolicyValueOnly() const
{
    return m_computedStyle.hasExplicitlySetTestRenderStyleHasExplicitlySetPolicyValueOnly();
}

inline Style::TestEnumeration RenderStyleProperties::testRenderStyleStorageOneLevelEnum() const
{
    return m_computedStyle.testRenderStyleStorageOneLevelEnum();
}

inline Style::TestRaw RenderStyleProperties::testRenderStyleStorageOneLevelRaw() const
{
    return m_computedStyle.testRenderStyleStorageOneLevelRaw();
}

inline const Style::Number<>& RenderStyleProperties::testRenderStyleStorageOneLevelReference() const
{
    return m_computedStyle.testRenderStyleStorageOneLevelReference();
}

inline Style::Number<> RenderStyleProperties::testRenderStyleStorageOneLevelValue() const
{
    return m_computedStyle.testRenderStyleStorageOneLevelValue();
}

inline Style::TestEnumeration RenderStyleProperties::testRenderStyleStorageTwoLevelEnum() const
{
    return m_computedStyle.testRenderStyleStorageTwoLevelEnum();
}

inline Style::TestRaw RenderStyleProperties::testRenderStyleStorageTwoLevelRaw() const
{
    return m_computedStyle.testRenderStyleStorageTwoLevelRaw();
}

inline const Style::Number<>& RenderStyleProperties::testRenderStyleStorageTwoLevelReference() const
{
    return m_computedStyle.testRenderStyleStorageTwoLevelReference();
}

inline Style::Number<> RenderStyleProperties::testRenderStyleStorageTwoLevelValue() const
{
    return m_computedStyle.testRenderStyleStorageTwoLevelValue();
}

inline Style::Number<> RenderStyleProperties::testLogicalPropertyGroupPhysicalHorizontal() const
{
    return m_computedStyle.testLogicalPropertyGroupPhysicalHorizontal();
}

inline Style::Number<> RenderStyleProperties::testLogicalPropertyGroupPhysicalVertical() const
{
    return m_computedStyle.testLogicalPropertyGroupPhysicalVertical();
}

inline Style::Number<> RenderStyleProperties::logicalTestLogicalPropertyGroupPhysicalHorizontal(WritingMode writingMode) const
{
    return m_computedStyle.logicalTestLogicalPropertyGroupPhysicalHorizontal(writingMode);
}

inline Style::Number<> RenderStyleProperties::logicalTestLogicalPropertyGroupPhysicalVertical(WritingMode writingMode) const
{
    return m_computedStyle.logicalTestLogicalPropertyGroupPhysicalVertical(writingMode);
}

inline Style::Number<> RenderStyleProperties::logicalTestLogicalPropertyGroupPhysicalHorizontal() const
{
    return m_computedStyle.logicalTestLogicalPropertyGroupPhysicalHorizontal();
}

inline Style::Number<> RenderStyleProperties::logicalTestLogicalPropertyGroupPhysicalVertical() const
{
    return m_computedStyle.logicalTestLogicalPropertyGroupPhysicalVertical();
}

} // namespace WebCore

